package com.sun.jndi.dns;

import javax.naming.InvalidNameException;
import java.nio.charset.StandardCharsets;
import java.io.IOException;
import javax.naming.CommunicationException;

public class ResourceRecord
{
    static final int TYPE_A = 1;
    static final int TYPE_NS = 2;
    static final int TYPE_CNAME = 5;
    static final int TYPE_SOA = 6;
    static final int TYPE_PTR = 12;
    static final int TYPE_HINFO = 13;
    static final int TYPE_MX = 15;
    static final int TYPE_TXT = 16;
    static final int TYPE_AAAA = 28;
    static final int TYPE_SRV = 33;
    static final int TYPE_NAPTR = 35;
    static final int QTYPE_AXFR = 252;
    static final int QTYPE_STAR = 255;
    static final String[] rrTypeNames;
    static final int CLASS_INTERNET = 1;
    static final int CLASS_HESIOD = 2;
    static final int QCLASS_STAR = 255;
    static final String[] rrClassNames;
    private static final int MAXIMUM_COMPRESSION_REFERENCES = 16;
    byte[] msg;
    int msgLen;
    boolean qSection;
    int offset;
    int rrlen;
    DnsName name;
    int rrtype;
    String rrtypeName;
    int rrclass;
    String rrclassName;
    int ttl;
    int rdlen;
    Object rdata;
    private static final boolean debug = false;
    
    ResourceRecord(byte[] msg, int msgLen, int offset, boolean qSection, boolean b) throws CommunicationException {
        this.ttl = 0;
        this.rdlen = 0;
        this.rdata = null;
        this.msg = msg;
        this.msgLen = msgLen;
        this.offset = offset;
        this.qSection = qSection;
        this.decode(b);
    }
    
    @Override
    public String toString() {
        String str = this.name + " " + this.rrclassName + " " + this.rrtypeName;
        if (!this.qSection) {
            str = str + " " + this.ttl + " " + ((this.rdata != null) ? this.rdata : "[n/a]");
            return str;
        }
        return str;
    }
    
    public DnsName getName() {
        return this.name;
    }
    
    public int size() {
        return this.rrlen;
    }
    
    public int getType() {
        return this.rrtype;
    }
    
    public int getRrclass() {
        return this.rrclass;
    }
    
    public Object getRdata() {
        return this.rdata;
    }
    
    public static String getTypeName(int n) {
        return valueToName(n, ResourceRecord.rrTypeNames);
    }
    
    public static int getType(String s) {
        return nameToValue(s, ResourceRecord.rrTypeNames);
    }
    
    public static String getRrclassName(int n) {
        return valueToName(n, ResourceRecord.rrClassNames);
    }
    
    public static int getRrclass(String s) {
        return nameToValue(s, ResourceRecord.rrClassNames);
    }
    
    private static String valueToName(int i, String[] array) {
        String string = null;
        if (i > 0 && i < array.length) {
            string = array[i];
        }
        else if (i == 255) {
            string = "*";
        }
        if (string == null) {
            string = Integer.toString(i);
            return string;
        }
        return string;
    }
    
    private static int nameToValue(String s, String[] array) {
        if (s.equals("")) {
            return -1;
        }
        if (s.equals("*")) {
            return 255;
        }
        if (Character.isDigit(s.charAt(0))) {
            try {
                return Integer.parseInt(s);
            }
            catch (NumberFormatException ex) {}
        }
        int i = 1;
        while (i < array.length) {
            if (array[i] != null && s.equalsIgnoreCase(array[i])) {
                return i;
            }
            ++i;
        }
        return -1;
    }
    
    public static int compareSerialNumbers(long n, long n2) {
        long n3 = n2 - n;
        if (n3 == 0L) {
            return 0;
        }
        if ((n3 > 0L && n3 <= 2147483647L) || (n3 < 0L && -n3 > 2147483647L)) {
            return -1;
        }
        return 1;
    }
    
    private void decode(boolean b) throws CommunicationException {
        int offset = this.offset;
        this.name = new DnsName();
        int decodeName = this.decodeName(offset, this.name);
        this.rrtype = this.getUShort(decodeName);
        this.rrtypeName = ((this.rrtype < ResourceRecord.rrTypeNames.length) ? ResourceRecord.rrTypeNames[this.rrtype] : null);
        if (this.rrtypeName == null) {
            this.rrtypeName = Integer.toString(this.rrtype);
        }
        decodeName += 2;
        this.rrclass = this.getUShort(decodeName);
        this.rrclassName = ((this.rrclass < ResourceRecord.rrClassNames.length) ? ResourceRecord.rrClassNames[this.rrclass] : null);
        if (this.rrclassName == null) {
            this.rrclassName = Integer.toString(this.rrclass);
        }
        decodeName += 2;
        if (!this.qSection) {
            this.ttl = this.getInt(decodeName);
            decodeName += 4;
            this.rdlen = this.getUShort(decodeName);
            decodeName += 2;
            this.rdata = ((b || this.rrtype == 6) ? this.decodeRdata(decodeName) : null);
            if ((this.rdata instanceof DnsName)) {
                this.rdata = this.rdata.toString();
            }
            decodeName += this.rdlen;
        }
        this.rrlen = decodeName - this.offset;
        this.msg = null;
    }
    
    private int getUByte(int n) {
        return this.msg[n] & 0xFF;
    }
    
    private int getUShort(int n) {
        return (this.msg[n] & 0xFF) << 8 | (this.msg[n + 1] & 0xFF);
    }
    
    private int getInt(int n) {
        return this.getUShort(n) << 16 | this.getUShort(n + 2);
    }
    
    private long getUInt(int n) {
        return (long)this.getInt(n) & 0xFFFFFFFFL;
    }
    
    private DnsName decodeName(int n) throws CommunicationException {
        DnsName dnsName = new DnsName();
        this.decodeName(n, dnsName);
        return dnsName;
    }
    
    private int decodeName(int offset, DnsName dnsName) throws CommunicationException {
        int n = -1;
        int i = 0;
        Label_0192: {
            try {
                while (i <= 16) {
                    int n2 = this.msg[offset] & 0xFF;
                    if (n2 == 0) {
                        ++offset;
                        dnsName.add(0, "");
                        break Label_0192;
                    }
                    if (n2 <= 63) {
                        ++offset;
                        dnsName.add(0, new String(this.msg, offset, n2, StandardCharsets.ISO_8859_1));
                        offset += n2;
                    }
                    else if ((n2 & 0xC0) == 0xC0) {
                        ++i;
                        if (n == -1) {
                            n = offset + 2;
                        }
                        offset = (this.getUShort(offset) & 0x3FFF);
                    }
                    else {
                        throw new IOException("Invalid label type: " + n2);
                    }
                }
                throw new IOException("Too many compression references");
            }
            catch (IOException | InvalidNameException cause) {
                CommunicationException ex = new CommunicationException("DNS error: malformed packet");
                ex.initCause(cause);
                throw ex;
            }
        }
        if (n == -1) {
            n = offset;
        }
        return n;
    }
    
    private Object decodeRdata(int n) throws CommunicationException {
        if (this.rrclass == 1) {
            switch (this.rrtype) {
                case 1: {
                    return this.decodeA(n);
                }
                case 28: {
                    return this.decodeAAAA(n);
                }
                case 2:
                case 5:
                case 12: {
                    return this.decodeName(n);
                }
                case 15: {
                    return this.decodeMx(n);
                }
                case 6: {
                    return this.decodeSoa(n);
                }
                case 33: {
                    return this.decodeSrv(n);
                }
                case 35: {
                    return this.decodeNaptr(n);
                }
                case 16: {
                    return this.decodeTxt(n);
                }
                case 13: {
                    return this.decodeHinfo(n);
                }
                default: {
                    break;
                }
            }
        }
        byte[] array = new byte[this.rdlen];
        System.arraycopy(this.msg, n, array, 0, this.rdlen);
        return array;
    }
    
    private String decodeMx(int n) throws CommunicationException {
        int uShort = this.getUShort(n);
        n += 2;
        return uShort + " " + this.decodeName(n);
    }
    
    private String decodeSoa(int n) throws CommunicationException {
        DnsName obj = new DnsName();
        n = this.decodeName(n, obj);
        DnsName obj2 = new DnsName();
        n = this.decodeName(n, obj2);
        long uInt = this.getUInt(n);
        n += 4;
        long uInt2 = this.getUInt(n);
        n += 4;
        long uInt3 = this.getUInt(n);
        n += 4;
        long uInt4 = this.getUInt(n);
        n += 4;
        long uInt5 = this.getUInt(n);
        n += 4;
        return obj + " " + obj2 + " " + uInt + " " + uInt2 + " " + uInt3 + " " + uInt4 + " " + uInt5;
    }
    
    private String decodeSrv(int n) throws CommunicationException {
        int uShort = this.getUShort(n);
        n += 2;
        int uShort2 = this.getUShort(n);
        n += 2;
        int uShort3 = this.getUShort(n);
        n += 2;
        return uShort + " " + uShort2 + " " + uShort3 + " " + this.decodeName(n);
    }
    
    private String decodeNaptr(int n) throws CommunicationException {
        int uShort = this.getUShort(n);
        n += 2;
        int uShort2 = this.getUShort(n);
        n += 2;
        StringBuffer obj = new StringBuffer();
        n += this.decodeCharString(n, obj);
        StringBuffer obj2 = new StringBuffer();
        n += this.decodeCharString(n, obj2);
        StringBuffer obj3 = new StringBuffer(this.rdlen);
        n += this.decodeCharString(n, obj3);
        return uShort + " " + uShort2 + " " + obj + " " + obj2 + " " + obj3 + " " + this.decodeName(n);
    }
    
    private String decodeTxt(int i) {
        StringBuffer sb = new StringBuffer(this.rdlen);
        int n = i + this.rdlen;
        while (i < n) {
            i += this.decodeCharString(i, sb);
            if (i >= n) {
                continue;
            }
            sb.append(' ');
        }
        return sb.toString();
    }
    
    private String decodeHinfo(int n) {
        StringBuffer sb = new StringBuffer(this.rdlen);
        n += this.decodeCharString(n, sb);
        sb.append(' ');
        n += this.decodeCharString(n, sb);
        return sb.toString();
    }
    
    private int decodeCharString(int n, StringBuffer sb) {
        int length = sb.length();
        int uByte = this.getUByte(n++);
        int n2 = uByte == 0;
        for (int i = 0; i < uByte; ++i) {
            int uByte2 = this.getUByte(n++);
            n2 |= (uByte2 == 32);
            if (uByte2 == 92 || uByte2 == 34) {
                n2 = true;
                sb.append('\\');
            }
            sb.append((char)uByte2);
        }
        if (n2 != 0) {
            sb.insert(length, '\"');
            sb.append('\"');
        }
        return uByte + 1;
    }
    
    private String decodeA(int n) {
        return (this.msg[n] & 0xFF) + "." + (this.msg[n + 1] & 0xFF) + "." + (this.msg[n + 2] & 0xFF) + "." + (this.msg[n + 3] & 0xFF);
    }
    
    private String decodeAAAA(int n) {
        int[] array = new int[8];
        for (int i = 0; i < 8; ++i) {
            array[i] = this.getUShort(n);
            n += 2;
        }
        int n2 = -1;
        int n3 = 0;
        int n4 = -1;
        int n5 = 0;
        for (int j = 0; j < 8; ++j) {
            if (array[j] == 0) {
                if (n2 == -1) {
                    n2 = j;
                    n3 = 1;
                }
                else if (++n3 >= 2 && n3 > n5) {
                    n4 = n2;
                    n5 = n3;
                }
            }
            else {
                n2 = -1;
            }
        }
        if (n4 == 0) {
            if (n5 == 6 || (n5 == 7 && array[7] > 1)) {
                return "::" + this.decodeA(n - 4);
            }
            if (n5 == 5 && array[5] == 65535) {
                return "::ffff:" + this.decodeA(n - 4);
            }
        }
        int n6 = n4 != -1;
        StringBuffer o = new StringBuffer(40);
        if (n4 == 0) {
            o.append(':');
        }
        for (int k = 0; k < 8; ++k) {
            if (n6 == 0 || k < n4 || k >= n4 + n5) {
                o.append(Integer.toHexString(array[k]));
                if (k < 7) {
                    o.append(':');
                }
            }
            else if (n6 && k == n4) {
                o.append(':');
            }
        }
        return o.toString();
    }
    
    private static void dprint(String s) {
    }
    
    static {
        rrTypeNames = new String[] { null, "A", "NS", null, null, "CNAME", "SOA", null, null, null, null, null, "PTR", "HINFO", null, "MX", "TXT", null, null, null, null, null, null, null, null, null, null, null, "AAAA", null, null, null, null, "SRV", null, "NAPTR" };
        rrClassNames = new String[] { null, "IN", null, null, "HS" };
    }
}
